@@ -5,6 +5,8 @@ import dash_core_components as dcc
 import dash_cytoscape as cyto
 import dash_html_components as html
 import networkx as nx
+import visdcc
+
 from dash.dependencies import ClientsideFunction
 from dash.exceptions import PreventUpdate
 
@@ -14,10 +16,11 @@ from .LayoutHelper import (
     create_edges_cytoscape,
     create_monitor_graph,
     create_nodes_cytoscape,
+    extract_param_dropdown,
+    get_param_dash_component,
 )
 from .. import agents as agentmet4fof_module
 
-
 class Dashboard_agt_net(Dashboard_Layout_Base):
     def set_layout_name(self, id="agt-net", title="Agent Network"):
         self.id = id
@@ -27,7 +30,7 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
         return [dcc.Graph(id='monitors-graph-' + str(i), figure={}, style={'height': '90vh'}) for i in
                 range(num_monitors)]
 
-    def get_layout(self, update_interval_seconds=3, num_monitors=10):
+    def get_layout(self):
         # body
         return html.Div(className="row", children=[
 
@@ -48,6 +51,9 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
                             html.Div(className="col", children=[
                                 LayoutHelper.html_button(icon="restore", text="Reset", id="reset-button")
                             ]),
+                            html.Div(className="col", children=[
+                                LayoutHelper.html_button(icon="restore", text="Export JPG", id="cyto-button-export")
+                            ]),
                         ])
 
                     ]),
@@ -55,59 +61,9 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
                         cyto.Cytoscape(
                             id='agents-network',
                             layout={'name': 'breadthfirst'},
-                            style={'width': '100%', 'height': '600px'},
+                            style={'width': '100%', 'height': '800px'},
                             elements=[],
-                            stylesheet=[
-                                {'selector': 'node', 'style':
-                                    {'label': 'data(id)',
-                                     'shape': 'rectangle'}
-                                 },
-                                {'selector': 'edge',
-                                 'style': {'curve-style': 'bezier', 'mid-target-arrow-shape':
-                                     'triangle', 'arrow-scale': 2, 'line-color': '#4287f5',
-                                           'mid-target-arrow-color': '#4287f5'},
-                                 },
-                                {
-                                    'selector': '.rectangle',
-                                    'style': {
-                                        'shape': 'rectangle'
-                                    }
-                                },
-                                {
-                                    'selector': '.triangle',
-                                    'style': {
-                                        'shape': 'triangle'
-                                    }
-                                },
-                                {
-                                    'selector': '.octagon',
-                                    'style': {
-                                        'shape': 'octagon'
-                                    }
-                                },
-                                {
-                                    'selector': '.ellipse',
-                                    'style': {
-                                        'shape': 'ellipse'
-                                    }
-                                },
-                                {
-                                    'selector': '.bluebackground',
-                                    'style': {
-                                        'background-color': '#c4fdff'
-                                    }
-                                },
-                                {
-                                    'selector': '.blue',
-                                    'style': {
-                                        'background-color': '#006db5'
-                                    }
-                                },
-                                {
-                                    'selector': '.coalition',
-                                    'style': {'line-style': 'dashed'}
-                                },
-                            ]
+                            stylesheet=self.app.network_stylesheet,
 
                         )
 
@@ -116,7 +72,7 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
                 ]),
                 html.H5(className="card", id="matplotlib-division", children=" "),
                 html.Div(className="card", id="monitors-temp-division",
-                         children=self.get_multiple_graphs(num_monitors)),
+                         children=self.get_multiple_graphs(self.app.num_monitors)),
 
             ]),
 
@@ -127,18 +83,41 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
 
                         html.Div(style={'margin-top': '20px'}, children=[
                             html.H6(className="black-text", children="Add Agent"),
-                            dcc.Dropdown(id="add-modules-dropdown"),
+                            dcc.Dropdown(id="add-modules-dropdown", style={'margin-bottom': '35px'}),
+
+                            html.Div(className="input-field", id="agent-init-div", children=[
+                                html.I(className="material-icons prefix", children=["account_circle"]),
+                                dcc.Input(
+                                    id="agent-init-name",
+                                    type="text",
+                                    value="",
+                                    className="validate"
+                                ),
+                                html.Label(children="New Agent Name", htmlFor="agent-init-name", className="active")
+                            ]),
+                            html.Div(style={'margin-top': '10px'}, id="agent-init-params", children=[
+
+                            ]),
                             LayoutHelper.html_button(icon="person_add", text="Add Agent", id="add-module-button"),
                             LayoutHelper.html_button(icon="delete_forever", text="Remove Agent",
                                                      id="remove-module-button", style={"margin-left": '4px'})
 
                         ]),
 
-                        html.Div(style={'margin-top': '20px'}, children=[
-                            html.H6(className="black-text", children="Dataset"),
-                            dcc.Dropdown(id="add-dataset-dropdown"),
-                            LayoutHelper.html_button(icon="add_to_queue", text="Add Datastream Agent",
-                                                     id="add-dataset-button")
+                        html.Div(style={'margin-top': '30px'}, children=[
+                            html.H6(className="black-text", children="Add Coalition", style={'margin-bottom': '20px'}),
+                            html.Div(className="input-field", children=[
+                                html.I(className="material-icons prefix", children=["dvr"]),
+                                dcc.Input(
+                                    id="coalition-name",
+                                    type="text",
+                                    value="",
+                                    className="validate"
+                                ),
+                                html.Label(children="New Coalition Name", htmlFor="coalition-name")
+                            ]),
+                            LayoutHelper.html_button(icon="add_to_queue", text="Add Coalition",
+                                                     id="add-coalition-button")
                         ])
                     ])
                 ]),
@@ -148,7 +127,7 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
                     html.Div(className="card-content", children=[
                         # side panel contents here
                         html.H5("Agent Configuration"),
-                        html.H5(id='selected-node', children="Not selected", className="flow-text",
+                        html.H6(id='selected-node', children="Not selected", className="flow-text",
                                 style={"font-weight": "bold"}),
 
                         html.H6(id='input-names', children="Not selected", className="flow-text"),
@@ -168,6 +147,7 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
                         html.P(id="disconnect_placeholder", className="black-text", children=" "),
                         html.P(id="monitor_placeholder", className="black-text", children=" "),
                         html.P(id="mpld3_placeholder", className="black-text", children=" "),
+                        visdcc.Run_js(id='toast-js-script')
                     ])
 
                 ])
@@ -175,7 +155,7 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
             ]),
             dcc.Interval(
                 id='interval-component-network-graph',
-                interval=update_interval_seconds * 1000,  # in milliseconds
+                interval=self.app.update_interval_seconds * 1000,  # in milliseconds
                 n_intervals=0
             ),
             dcc.Interval(
@@ -185,19 +165,40 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
             ),
             dcc.Interval(
                 id='interval-update-monitor-graph',
-                interval=update_interval_seconds * 1000,  # in milliseconds
+                interval=self.app.update_interval_seconds * 1000,  # in milliseconds
+                n_intervals=0
+            ),
+            dcc.Interval(
+                id='interval-update-toast',
+                interval=1 * 1000,  # in milliseconds
                 n_intervals=0
             )
         ])
 
     def prepare_callbacks(self, app):
+
+        @app.callback([dash.dependencies.Output('agents-network', 'generateImage')],
+                       [dash.dependencies.Input('cyto-button-export', 'n_clicks')],
+                      )
+        def export_image(n_clicks):
+            if (n_clicks is not None):
+                return [{
+                    'type': "png",
+                    'action': "download"
+                }]
+            else:
+                raise PreventUpdate
+
         # Update network graph per interval
         @app.callback([dash.dependencies.Output('agents-network', 'elements'),
                        dash.dependencies.Output('connect-modules-dropdown', 'options')],
-                      [dash.dependencies.Input('interval-component-network-graph', 'n_intervals')],
+                      [dash.dependencies.Input('interval-component-network-graph', 'n_intervals'),
+                       # dash.dependencies.Input('connect-module-button', 'n_clicks')
+                       ],
                       [dash.dependencies.State('agents-network', 'elements')]
                       )
         def update_network_graph(n_intervals, graph_elements):
+
             # get nameserver
             agentNetwork = app.dashboard_ctrl.agentNetwork
 
@@ -207,21 +208,24 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
 
             # get coalitions of agents
             coalitions = agentNetwork.coalitions
+            num_agent_coalitions = sum([len(coalition.agent_names()) for coalition in coalitions])
+            if not hasattr(app, "num_agent_coalitions"):
+                app.num_agent_coalitions = num_agent_coalitions
 
             # if current graph number is different more than before, then update graph
-            if len(graph_elements) != (len(nodes) + len(edges) + len(coalitions)):
+            if (len(graph_elements) != (len(nodes) + len(edges) + len(coalitions))) or (app.num_agent_coalitions != num_agent_coalitions):
                 # if(app.dashboard_ctrl.agent_graph.number_of_nodes() != len(nodes) or app.dashboard_ctrl.agent_graph.number_of_edges() != len(edges)) or n_intervals == 0:
                 new_G = nx.DiGraph()
                 new_G.add_nodes_from(nodes(data=True))
                 new_G.add_edges_from(edges)
 
                 nodes_elements = create_nodes_cytoscape(new_G)
-                edges_elements = create_edges_cytoscape(edges)
+                edges_elements = create_edges_cytoscape(edges, app.hide_default_edge)
                 # print(edges_elements)
                 # draw coalition nodes, and assign child nodes to coalition nodes
                 if len(agentNetwork.coalitions) > 0:
                     parent_elements = [{"data": {'id': coalition.name, 'label': coalition.name},
-                                        'classes': 'bluebackground'} for coalition in agentNetwork.coalitions]
+                                        'classes': 'coalition'} for coalition in agentNetwork.coalitions]
                     for coalition in coalitions:
                         # check if agent is in the coalition, set its parents
                         for agent_node in nodes_elements:
@@ -232,11 +236,13 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
                         for edges in edges_elements:
                             if edges["data"]["source"] in coalition.agent_names() and edges["data"][
                                 "target"] in coalition.agent_names():
-                                edges.update({'classes': "coalition"})
+                                edges.update({'classes': "coalition-edge"})
                 else:
                     parent_elements = []
 
+                # update agent graph and number of coalitions
                 app.dashboard_ctrl.agent_graph = new_G
+                app.num_agent_coalitions = num_agent_coalitions
 
                 # update agents connect options
                 node_connect_options = [{'label': agentName, 'value': agentName} for agentName in nodes]
@@ -247,23 +253,18 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
                 raise PreventUpdate
 
         @app.callback([dash.dependencies.Output('add-modules-dropdown', 'options'),
-                       dash.dependencies.Output('add-dataset-dropdown', 'options')
                        ],
                       [dash.dependencies.Input('interval-add-module-list', 'n_intervals')
                        ])
         def update_add_module_list(n_interval):
             # get nameserver
-            # agentNetwork = dashboard_ctrl.agentNetwork
-
-            # agentTypes = dashboard_ctrl.get_agentTypes()
-            # module_add_options = [ {'label': agentType, 'value': agentType} for agentType in list(agentTypes.keys())]
             agentTypes = app.dashboard_ctrl.get_agentTypes()
             module_add_options = [{'label': agentType, 'value': agentType} for agentType in list(agentTypes.keys())]
 
             datasets = app.dashboard_ctrl.get_datasets()
             module_dataset_options = [{'label': dataset, 'value': dataset} for dataset in list(datasets.keys())]
 
-            return [module_add_options, module_dataset_options]
+            return [module_add_options]
 
         # Start button click
         @app.callback(dash.dependencies.Output('start-button', 'children'),
@@ -272,6 +273,7 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
         def start_button_click(n_clicks):
             if n_clicks is not None:
                 app.dashboard_ctrl.agentNetwork.set_running_state()
+                raise_toast("Set agents state : %s !" % "Running")
             raise PreventUpdate
 
         # Stop button click
@@ -281,6 +283,7 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
         def stop_button_click(n_clicks):
             if n_clicks is not None:
                 app.dashboard_ctrl.agentNetwork.set_stop_state()
+                raise_toast("Set agents state : %s !" % "Stop")
             raise PreventUpdate
 
         # Handle click of the reset button.
@@ -291,18 +294,47 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
         def reset_button_click(n_clicks):
             if n_clicks is not None:
                 app.dashboard_ctrl.agentNetwork.reset_agents()
+                raise_toast("Reset agent states !")
             raise PreventUpdate
 
+
+        # Init Agent Parameters choices
+        @app.callback([dash.dependencies.Output('agent-init-name', 'value'),
+                       dash.dependencies.Output('agent-init-div', 'style'),
+                       dash.dependencies.Output('agent-init-params', 'children')],
+                      [dash.dependencies.Input('add-modules-dropdown', 'value')],
+                      )
+        def add_agent_init_params(add_dropdown_val):
+            # selected a class from the dropdown
+            agentNetwork = app.dashboard_ctrl.agentNetwork
+            if add_dropdown_val is not None:
+                selected_agent_class = app.dashboard_ctrl.get_agentTypes()[add_dropdown_val]
+                if hasattr(selected_agent_class,'parameter_choices'):
+                    init_param_components =  [get_param_dash_component(key,val) for key,val in selected_agent_class.parameter_choices.items()]
+                else:
+                    init_param_components = []
+
+                unique_agent_name =agentNetwork.generate_module_name_byType(selected_agent_class)
+
+                return [unique_agent_name, {"display":"block"}, init_param_components]
+            else:
+                return ["", {"display":"none"},[]]
+
         # Add agent button click
         @app.callback(dash.dependencies.Output('add-module-button', 'children'),
                       [dash.dependencies.Input('add-module-button', 'n_clicks')],
-                      [dash.dependencies.State('add-modules-dropdown', 'value')]
+                      [dash.dependencies.State('add-modules-dropdown', 'value'),
+                       dash.dependencies.State('agent-init-name', 'value'),
+                       dash.dependencies.State('agent-init-params', 'children'),
+                       ]
                       )
-        def add_agent_button_click(n_clicks, add_dropdown_val):
+        def add_agent_button_click(n_clicks, add_dropdown_val, init_name_input, init_params_div):
             # for add agent button click
             if n_clicks is not None:
                 agentTypes = app.dashboard_ctrl.get_agentTypes()
-                new_agent = app.dashboard_ctrl.agentNetwork.add_agent(agentType=agentTypes[add_dropdown_val])
+                init_params_kwargs = extract_param_dropdown(init_params_div)
+                new_agent = app.dashboard_ctrl.agentNetwork.add_agent(name=init_name_input, agentType=agentTypes[add_dropdown_val], **init_params_kwargs)
+                raise_toast("Spawned new agent : %s !" % init_name_input)
             raise PreventUpdate
 
         # Add agent button click
@@ -314,24 +346,23 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
             # for add agent button click
             if n_clicks is not None and current_agent_id != "Not selected":
                 app.dashboard_ctrl.agentNetwork.remove_agent(current_agent_id)
+                raise_toast("Removed agent : %s !" % current_agent_id)
             raise PreventUpdate
 
-        # Add agent button click
-        @app.callback(dash.dependencies.Output('add-dataset-button', 'children'),
-                      [dash.dependencies.Input('add-dataset-button', 'n_clicks')],
-                      [dash.dependencies.State('add-dataset-dropdown', 'value')]
+        # Add coalition button click
+        @app.callback([dash.dependencies.Output('coalition-name', 'value')],
+                      [dash.dependencies.Input('add-coalition-button', 'n_clicks')],
+                      [dash.dependencies.State('coalition-name', 'value'),
+                       ]
                       )
-        def add_dataset_button_click(n_clicks, add_dropdown_val):
+        def add_coalition_button_click(n_clicks, coalition_name):
             # for add agent button click
             if n_clicks is not None:
-                agentTypes = app.dashboard_ctrl.get_agentTypes()
-                datasets = app.dashboard_ctrl.get_datasets()
-                chosen_dataset = datasets[add_dropdown_val]()
-                new_agent = app.dashboard_ctrl.agentNetwork.add_agent(name=type(chosen_dataset).__name__,
-                                                                      agentType=agentmet4fof_module.ML_DataStreamAgent)
+                if coalition_name != "":
+                    new_coalition = app.dashboard_ctrl.agentNetwork.add_coalition(name=coalition_name)
+                    raise_toast("Created new coalition : %s !" % coalition_name)
+            return [""]
 
-                new_agent.init_parameters(stream=chosen_dataset)
-            raise PreventUpdate
 
         @app.callback(dash.dependencies.Output('connect_placeholder', 'children'),
                       [dash.dependencies.Input('connect-module-button', 'n_clicks')],
@@ -341,25 +372,44 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
             if n_clicks_connect is not None and current_agent_id != "Not selected":
                 # get nameserver
                 agentNetwork = app.dashboard_ctrl.agentNetwork
-
+                target_agent = agentNetwork.get_agent(dropdown_value)
                 # for connect module button click
-                if current_agent_id != dropdown_value:
-                    agentNetwork.bind_agents(agentNetwork.get_agent(current_agent_id),
-                                             agentNetwork.get_agent(dropdown_value))
+                # connect agents
+                if current_agent_id in agentNetwork.agents():
+                    if current_agent_id != dropdown_value:
+                        agentNetwork.bind_agents(agentNetwork.get_agent(current_agent_id),
+                                                 target_agent)
+                        raise_toast("Connected %s to %s !" % (current_agent_id, target_agent.name))
+                # otherwise, selected is a coalition
+                # add it into coalition
+                else:
+                    agentNetwork.add_coalition_agent(name=current_agent_id, agents=[target_agent])
+                    raise_toast("%s joined %s coalition !" % (target_agent.name, current_agent_id))
+
             raise PreventUpdate
 
         @app.callback(dash.dependencies.Output('disconnect_placeholder', 'children'),
                       [dash.dependencies.Input('disconnect-module-button', 'n_clicks')],
                       [dash.dependencies.State('connect-modules-dropdown', 'value'),
                        dash.dependencies.State('selected-node', 'children')])
-        def unbind_module_click(n_clicks_connect, dropdown_value, current_agent_id):
-            if (n_clicks_connect is not None):
+        def unbind_module_click(n_clicks_disconnect, dropdown_value, current_agent_id):
+            if (n_clicks_disconnect is not None):
                 # get nameserver
                 agentNetwork = app.dashboard_ctrl.agentNetwork
+                target_agent = agentNetwork.get_agent(dropdown_value)
 
                 # for connect module button click
-                agentNetwork.unbind_agents(agentNetwork.get_agent(current_agent_id),
-                                           agentNetwork.get_agent(dropdown_value))
+                # disconnect agents
+                if current_agent_id in agentNetwork.agents():
+                    if current_agent_id != dropdown_value:
+                        agentNetwork.unbind_agents(agentNetwork.get_agent(current_agent_id),
+                                                 target_agent)
+                        raise_toast("Disconnected %s from %s !" % (current_agent_id, target_agent.name))
+                # otherwise, selected is a coalition
+                # remove it from coalition
+                else:
+                    agentNetwork.remove_coalition_agent(coalition_name=current_agent_id, agent_name=target_agent.name)
+                    raise_toast("%s removed from %s coalition !" % (target_agent.name, current_agent_id))
             raise PreventUpdate
 
         @app.callback([dash.dependencies.Output('selected-node', 'children'),
@@ -378,19 +428,29 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
                 # get nameserver
                 agentNetwork = app.dashboard_ctrl.agentNetwork
 
-                app.dashboard_ctrl.current_selected_agent = agentNetwork.get_agent(current_agent_id)
-                input_names = list(app.dashboard_ctrl.current_selected_agent.get_attr('Inputs').keys())
-                output_names = list(app.dashboard_ctrl.current_selected_agent.get_attr('Outputs').keys())
-                agent_parameters = app.dashboard_ctrl.current_selected_agent.get_all_attr()
+                # selected agent
+                if current_agent_id in agentNetwork.agents():
+                    app.dashboard_ctrl.current_selected_agent = agentNetwork.get_agent(current_agent_id)
+                    input_names = list(app.dashboard_ctrl.current_selected_agent.get_attr('Inputs').keys())
+                    output_names = list(app.dashboard_ctrl.current_selected_agent.get_attr('Outputs').keys())
+                    agent_parameters = app.dashboard_ctrl.current_selected_agent.get_all_attr()
 
-                # formatting
-                input_names = ["Inputs: "] + [input_name + ", " for input_name in input_names]
-                output_names = ["Outputs: "] + [output_name + ", " for output_name in output_names]
-                agent_parameters_texts = [LayoutHelper.visualise_agent_parameters(k, v) for k, v in
-                                          agent_parameters.items()]
-                agent_parameters_div = [html.H6("Parameters: ")] + agent_parameters_texts
+                    # formatting
+                    input_names = ["Inputs: "] + [input_name + ", " for input_name in input_names]
+                    output_names = ["Outputs: "] + [output_name + ", " for output_name in output_names]
+                    agent_parameters_texts = [LayoutHelper.visualise_agent_parameters(k, v) for k, v in
+                                              agent_parameters.items()]
+                    agent_parameters_div = [html.H6("Parameters: ")] + agent_parameters_texts
+
+                    return [current_agent_id,  input_names, output_names, agent_parameters_div]
+
+                # selected coalition
+                else:
+                    coalition = agentNetwork.get_coalition(current_agent_id)
+                    agent_members = [html.H6("Agents: ")] + [str(coalition.agent_names())]
+
+                    return [current_agent_id,[], [], agent_members]
 
-                return [current_agent_id, input_names, output_names, agent_parameters_div]
             else:
                 return ["Not selected", input_names, output_names, agent_parameters_div]
 
@@ -411,7 +471,7 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
             if agentNetwork.get_mode() != "Running" and agentNetwork.get_mode() != "Reset" and n_interval > 0:
                 raise PreventUpdate
 
-            agent_names = agentNetwork.agents('MonitorAgent')  # get all agent names
+            agent_names = agentNetwork.agents(filter_agent='Monitor')  # get all agent names
             app.num_monitor = len(agent_names)
             monitor_graphs = [{'data': []} for i in range(app.num_monitors)]
             style_graphs = [{'opacity': 0, 'width': 10, 'height': 10} for i in range(app.num_monitors)]
@@ -444,6 +504,7 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
                             data.append(trace)
                     else:
                         data.append(traces)
+
                 if len(data) > 5:
                     y_title_offset = 0.1
                 else:
@@ -553,16 +614,16 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
             if agentNetwork.get_mode() != "Running" and n_interval > 0:
                 raise PreventUpdate
 
-            agent_names = agentNetwork.agents()  # get all agent names
-            agent_type = "Monitor"  # all agents with Monitor in its name will be selected
+            agent_name_filter = "Monitor"
+            agent_names = agentNetwork.agents(filter_agent=agent_name_filter)  # get all agent names
+              # all agents with Monitor in its name will be selected
             plots_data = {}  # storage for all monitor agent's memory
 
             # load data from all Monitor agent's memory
             for agent_name in agent_names:
-                if agent_type in agent_name:
-                    monitor_agent = agentNetwork.get_agent(agent_name)
-                    plots = monitor_agent.get_attr('plots')
-                    plots_data.update({agent_name: plots})
+                monitor_agent = agentNetwork.get_agent(agent_name)
+                plots = monitor_agent.get_attr('plots')
+                plots_data.update({agent_name: plots})
 
             # now monitors_data = {'monitor_agent1_name':agent1_memory, 'monitor_agent2_name':agent2_memory }
             # now create a plot for each monitor agent
@@ -599,6 +660,18 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
             # set dimensions of each monitor agent's graph
             return [all_graphs]
 
+        @app.callback(
+            dash.dependencies.Output('toast-js-script', 'run'),
+            [dash.dependencies.Input('interval-update-toast', 'n_intervals')])
+        def generate_toast_msg(n_intervals):
+            # if there are messages in toast contents to be displayed
+            if hasattr(app, "toast_contents") and len(app.toast_contents) > 0:
+                # pop toast contents
+                return "M.toast({html: '%s'})" % app.toast_contents.pop(0)
+
+            else:
+                return ""
+
         app.clientside_callback(
             ClientsideFunction(
                 namespace='clientside',
@@ -608,4 +681,9 @@ class Dashboard_agt_net(Dashboard_Layout_Base):
             [dash.dependencies.Input('matplotlib-division', 'children')]
         )
 
+        app.toast_contents = []
+
+        def raise_toast(message, app=app):
+            app.toast_contents.append(message)
+
         return app
